#include-once
;#include <Array.au3>
;#include <Common\Debug.au3>

#cs ================================================================================================
This file contain different Total Commander related variables and code
#ce ================================================================================================

#cs ================================================================================================
2022-05-20 - renamed _FileWriteReplace => _TCHFileWriteReplace
2022-05-06 - fixed $TCCMD not returned some commands, that have two sets (ex. cm_SrcCustomView)
2021-09-11 - _TCHChangeExt renamed to _TCHChangeFileExt
2021-09-11 - _TCHExtract* functions revised and optimized for speed
2021-09-03 - fixed getting mapped icon
2021-08-16 - removed deprecated code, refactoring
2021-07-15 - fixed _TCHCreateButton wrong params order
2021-06-30 - added _TCHGetFullPath
2021-06-09 - _TCHRestart now aware of 32/34 version
2021-03-16 - fully rewritten icon management, implemented centralized custom icons 
2020-06-21 - added $TC_ARCH global
2020-06-09 - some cleanup
2020-05-01 - removed deprecated _TCHLoadFileList
2020-03-12 - moved Map-dependent code to TCHelpersMap.au3
2020-03-12 - redone translation code without maps
2020-03-11 - improved temporary files/dirs handling, added autodelete on exit
2018-04-25 - Added map helpers
2018-03-05 - _TCH_GetScriptInfo support for compiled scripts
2018-02-00 - Updated to latest changes in structure
2017-12-17 - Updated to latest changes in pack structure, fixed continious lists redirection awareness
2017-10-11 - Added temporary files handling functions
2017-03-21 - Fixed (again) _TCExtractFileName
2017-03-21 - Fixed wrong icon index in _TCHCreateButton function
2017-02-27 - Added history block, added _TCHSendMsg function, added _TCHLockWindowUpdate function
#ce ================================================================================================

; #CONSTANTS# ======================================================================================
; Misc globals
Global Const $sIniError = "<error>"

Global Const $TC_REGKEY_SYSTEM = "HKEY_LOCAL_MACHINE\Software\Ghisler\Total Commander"
Global Const $TC_REGKEY_USER = "HKEY_CURRENT_USER\Software\Ghisler\Total Commander"
Global Const $TC_SENDMESSAGE = 0x400 + 51 ; $WM_USER + 51
Global Const $TC_INFOREQUEST = 0x400 + 50 ; $WM_USER + 50

; Environment variables, that can be set at TC start with Autorun
; COMMANDER_LANG - use this language, autodetect intead
; COMMANDER_HWND - main window handle of current instance of TC
; COMMANDER_ARCH - architecture of current instance of TC

Global Const $TC_INI          = EnvGet("COMMANDER_INI") ; set by TC
Global Const $TC_PATH         = __TCHGetPath()
Global Const $TC_HWND         = __TCHGetActiveInstance()
Global Const $TC_LANG         = __TCHGetLangID() ; maybe on demand?
Global Const $TC_ARCH         = __TCHGetArch()
Global Const $TC_SCRIPTS      = $TC_PATH & "\Scripts"
Global Const $TC_PACKAGE      = $TC_PATH & "\Package"
Global Const $TC_ICONS        = $TC_PATH & "\Icons"
Global Const $TC_CONFIG       = $TC_PATH & "\Profile\Config"
Global Const $TC_CONFIG_LOCAL = $TC_PATH & "\Profile\Local"
Global Const $TC_TEMPDIR      = $TC_PATH & "\Profile\Temp"
Global Const $TC_LANGDIR      = $TC_PATH & "\Language"

Global Enum $IT_CUSTOM = 2, $IT_MAINLIB, $IT_MAINLIBDEF

; Pseudo-objects, incapsulating language, options and commands handling
Global Const $LANG   = _Lang
Global Const $OPTION = _Option
Global Const $TCCMD  = _IntCmd

; #INITIALIZATION# =================================================================================
_Option("CustomIconLib",      $TC_ICONS & "\Custom.icl")
_Option("IconMapperConfig",   $TC_ICONS & "\Mapper.ini")
_Option("SourcePath",         $TC_PATH & "\Scripts\Sources") ; Will be moved?
_Option("BarDataID",          "TOTALCMD#BAR#DATA")
; slower options are on demand, value retrieved at first call
_Option("TypesIconLib",       __TCHGetTypesLib)

If $TC_LANG = "RUS" Then
    _IntMsg("Message",        "")
    _IntMsg("NoInstance",     "Total Commander     .")
    _IntMsg("RunError",       "   Total Commander!")
    _IntMsg("NotEnoughParam", " %d  ,  : %d")
    _IntMsg("NoGetPath",      "     Total Commander.     Total Commander.")
    _IntMsg("ErrRemoveTemp",  "    /:")
Else
    _IntMsg("Message",        "Message")
    _IntMsg("NoInstance",     "Total Commander is not running or error occured.")
    _IntMsg("RunError",       "Can't run Total Commander!")
    _IntMsg("NotEnoughParam", "Passed %d parameters, should be: %d")
    _IntMsg("NoGetPath",      "Can't get the path to Total Commander. Try to run the program from inside of Total Commander.")
    _IntMsg("ErrRemoveTemp",  "Can't remove temporary files/dirs:")
Endif

; =============================================================================
; Option pseudo-object handling
; =============================================================================
; Handle values pairs for options
Func _Option($vID, $vValue = Default)
    Local Static $aList[0][2]
    ; we accept only one function GetList
    If IsFunc($vID) Then Return $vID($aList)
    Local $nIndex = _PairListKeyIndex($aList, $vID)
    If $nIndex = -1 Then
        ; add key/value
        Return $aList[ _PairListAdd($aList, $vID, $vValue) ][1]
    EndIf
    ; get value
    If $vValue = Default Then 
        Local $vRet = $aList[$nIndex][1]
        ; check if it macro value
        If IsFunc($vRet) Then Return $vRet()
        Return $vRet
    EndIf
    ; set value, return old value
    Local $vOld = $aList[$nIndex][1]
    $aList[$nIndex][1] = $vValue
    Return $vOld
EndFunc

; =============================================================================
; Language pseudo-object handling
; =============================================================================
; User language pseudo-object
Func _Lang($vID, $vParam1 = Default)
    Local Static $aList[0][2]
    Local Static $sLangFile = Default
    Return _LangHandle($aList, $sLangFile, $vID, $vParam1)
EndFunc

; Internal language pseudo-object
Func _IntMsg($vID, $vParam1 = Default)
    Local Static $aList[0][2]
    Local Static $sLangFile = Default
    Return _LangHandle($aList, $sLangFile, $vID, $vParam1)
EndFunc

; Implementation of handling language pseudo-objects
Func _LangHandle(ByRef $aList, ByRef $sLangFile, $vID, $vParam1 = Default)
    If IsFunc($vID) Then
        Switch FuncName($vID)
            Case "GetList"
                Return $vID($aList)
            Case "LangInit", "LangSetFile"
                Return $vID($aList, $sLangFile, $vParam1)
            Case "LangGetFile"
                Return $vID($sLangFile)
        EndSwitch
    EndIf
    Local $nIndex = _PairListKeyIndex($aList, $vID)
    If $nIndex <> -1 Then 
        ; Allow to redefine UDF-defined language strings
        If $vParam1 <> Default Then $aList[$nIndex][1] = $vParam1 
        Return $aList[$nIndex][1]
    EndIf 
    If $vParam1 = Default Then Return StringFormat("Missed: %s", $vID)
    Return $aList[ _PairListAdd($aList, $vID, $vParam1) ][1]
EndFunc

; Function for setting arbitrary language file 
Func LangSetFile(ByRef $aList, ByRef $sLangFile, $vParam1 = Default)
    If $vParam1 = Default Then 
        $sLangFile = StringLeft(@ScriptFullPath, StringInStr(@ScriptFullPath, ".", 0, -1)-1) & ".lng"
    Else
        $sLangFile = $vParam1
    EndIf
EndFunc

Func LangGetFile(ByRef $sLangFile)
    Return $sLangFile
EndFunc

; Function for initialization 
Func LangInit(ByRef $aList, ByRef $sLangFile, $sLangID = Default)
    If $sLangFile = Default Then LangSetFile($aList, $sLangFile)
    If $sLangID = Default Then $sLangID = $TC_LANG ; if no overriding then use
    If FileExists($sLangFile) Then
        For $i = 0 To UBound($aList)-1
            $aList[$i][1] = IniRead($sLangFile, $sLangID, $aList[$i][0], $aList[$i][1])
        Next
    EndIf
EndFunc

; function for retrieve full list
Func GetList(ByRef $aList)
    Return $aList
EndFunc

; Get key index
Func _PairListKeyIndex(ByRef $aList, $sName)
    For $i = 0 To UBound($aList) - 1
        If $sName = $aList[$i][0] Then Return $i
    Next
    Return -1
EndFunc

; Adds key/value pair
Func _PairListAdd(ByRef $aList, $sName, $vValue = Default)
    Local $nIndex = UBound($aList)
    ReDim $aList[$nIndex+1][2]
    $aList[$nIndex][0] = $sName
    $aList[$nIndex][1] = $vValue
    Return $nIndex
EndFunc

; =============================================================================
; Internal commands pseudo-object handling
; Commands inc-file loading at first call
; Performance: ~15ms first call, 0.06ms subsequent calls, ~40ms getting full list
; =============================================================================
Func _IntCmd($vID, $vParam1 = Default)
    Local Static $aList[0][2]
    If UBound($aList) = 0 Then _IntCmdLoad($aList)
    If IsFunc($vID) Then
        If FuncName($vID) <> "GetList" Then Return 0
        ; GetList return full commands list
        Local $aTempList = $aList, $nBound
        Local $aMCList = _GetIndexedCommand()
        For $i = 0 To UBound($aMCList)-1
            If $aMCList[$i][1] = 0 Then ContinueLoop
            $nBound = UBound($aTempList)
            ReDim $aTempList[ $nBound + $aMCList[$i][2] - $aMCList[$i][1] + 1 ][2]
            For $j = 0 To $aMCList[$i][2] - $aMCList[$i][1]
                $aTempList[ $nBound + $j ][0] = $aMCList[$i][0] & ($j + $aMCList[$i][3]) 
                $aTempList[ $nBound + $j ][1] = $aMCList[$i][1] + $j
            Next
        Next
        Return $vID($aTempList)
    EndIf 
    ; try to get indexed command
    Local $nCommand = 0
    Local $aCmd = StringRegExp($vID, "(?s)^([A-Za-z_]+?)(\d+?)$", 1)
    If IsArray($aCmd) Then $nCommand = _GetIndexedCommand($aCmd[0], $aCmd[1])
    ; if not multicommand, lookup in remain commands
    If $nCommand > 0 Then Return $nCommand
    Local $nIndex = _PairListKeyIndex($aList, $vID)
    If $nIndex = -1 Then Return 0
    Return $aList[$nIndex][1]
EndFunc

Func _IntCmdLoad(ByRef $aList)
    ; Performance impact about 15ms
    Local $sIncFile = EnvGet("COMMANDER_PATH") & "\TOTALCMD.INC"
    Local $aFile = StringRegExp(FileRead($sIncFile), "(cm_[a-zA-Z0-9]+?)=(\d+?);.+?\R", 4)
    If UBound($aFile) = 0 Then Return
    ; Redim to max items
    ReDim $aList[ UBound($aFile) ][2]

    Local $nCmdCount = -1, $aTemp, $aCmdName
    Local $aMultiCmd = _GetIndexedCommand()
    For $i = 0 To UBound($aFile)-1
        $aTemp = $aFile[$i]
      
        $aCmdName = StringRegExp($aTemp[1], "([a-zA-Z_]+)\d+$", 1)
        If @error = 0 Then
            For $j = 0 To UBound($aMultiCmd)-1
                If $aCmdName[0] = $aMultiCmd[$j][0] Then ContinueLoop(2) ; $aTemp[1]
            Next
        EndIf
        
        $nCmdCount += 1
        $aList[$nCmdCount][0] = $aTemp[1]
        $aList[$nCmdCount][1] = $aTemp[2]
    Next
    Local $nDrive = 2061
    For $i = Asc("A") To Asc("Z")
        $nCmdCount += 1
        $aList[$nCmdCount][0] = "cm_GotoDrive" & Chr($i)
        $aList[$nCmdCount][1] = $nDrive
        $nDrive += 1
    Next
    ReDim $aList[$nCmdCount+1][2]
EndFunc

Func _GetIndexedCommand($sCommand = Default, $nIndex = Default)
    ; Name, StartCmdValue, EndCmdValue, NameStartIndex, SearchNextSet
    Local Static $aCmdMap[][5] = [ _
        ["cm_UserMenu", 701, 799, 1] ,          ["cm_GotoDrive", 0, 0], _
        ["cm_SrcActivateTab", 5001, 5099, 1],   ["cm_TrgActivateTab", 5101, 5199, 1], _
        ["cm_LeftActivateTab", 5201, 5299, 1],  ["cm_RightActivateTab", 5301, 5399, 1], _
        ["cm_SrcSortByCol", 6001, 6099, 1],     ["cm_TrgSortByCol", 6101, 6199, 1], _
        ["cm_LeftSortByCol", 6201, 6299, 1],    ["cm_RightSortByCol", 6301, 6399, 1], _
        ["cm_SrcCustomView", 271, 299, 1, 1],   ["cm_SrcCustomView", 7530, 7599, 30], _
        ["cm_TrgCustomView", 421, 449, 1], _
        ["cm_LeftCustomView", 71, 99, 1, 1],    ["cm_LeftCustomView", 6530, 6599, 30], _
        ["cm_RightCustomView", 171, 199, 1, 1], ["cm_RightCustomView", 7030, 7099, 30], _
        ["cm_SrcViewMode", 8500, 8749, 0],      ["cm_TrgViewMode", 8750, 8999, 0], _
        ["cm_LeftViewMode", 8000, 8249, 0],     ["cm_RightViewMode", 8250, 8499, 0]]
    If $sCommand = Default Then Return $aCmdMap

    For $i = 0 To UBound($aCmdMap)-1
        If $aCmdMap[$i][0] = $sCommand Then 
            If $nIndex < $aCmdMap[$i][3] Or $nIndex > $aCmdMap[$i][2] - $aCmdMap[$i][1] + $aCmdMap[$i][3] Then 
                If $aCmdMap[$i][4] Then
                    ContinueLoop
                Else
                    Return 0
                EndIf
            EndIf
            Return $aCmdMap[$i][1] + ($nIndex - $aCmdMap[$i][3])
        EndIf 
    Next
    Return 0
EndFunc


; =============================================================================
; Temporary file/directory handling
; =============================================================================
; Function for keep track temp resources (does not required globals)
; Flags:
; 0 - just get unique name
; 1 - create directory with that name
; 2 - don't remove resource at exit
; prefix - arbitrary text before random string
; suffix - arbitrary text after random string (extension can be defined this way)
Func _TCHGetTempResource($nFlags, $sPrefix = "", $sSuffix = "")
    Local Static $aTempList[0][2]

    If $nFlags = -1 Then
        Local $sList = ""
        For $i = 0 To UBound($aTempList)-1
            ; file can be not created, skip
            If Not FileExists($aTempList[$i][0]) Then ContinueLoop
            If $aTempList[$i][1] = 1 Then
                FileChangeDir(@ScriptDir) ; get out of temp dir
                If Not DirRemove($aTempList[$i][0], 1) Then $sList &= $aTempList[$i][0] & @CRLF
            Else
                If Not FileDelete($aTempList[$i][0]) Then $sList &= $aTempList[$i][0] & @CRLF
            EndIf
        Next
        If $sList <> "" Then MsgBox(48, "", _IntMsg("ErrRemoveTemp") & @CRLF & $sList)
        Return
    EndIf
    
    Local $sTemp
    Do
        $sTemp = @TempDir & "\" & $sPrefix & _TCHGetRandomString(8) & $sSuffix
    Until Not FileExists($sTemp)

    If BitAND($nFlags, 1) Then
        If Not DirCreate($sTemp) Then Return SetError(1, 0, "")
    EndIf

    If Not BitAND($nFlags, 2) Then
        ReDim $aTempList[UBound($aTempList)+1][2]
        $aTempList[UBound($aTempList)-1][0] = $sTemp
        $aTempList[UBound($aTempList)-1][1] = BitAND($nFlags, 1)
        OnAutoItExitRegister("_TCHGetTempResourceOnExit")
    EndIf

    Return $sTemp
EndFunc

; OnAutoItExitRegister can only register function without params
Func _TCHGetTempResourceOnExit()
    _TCHGetTempResource(-1)
EndFunc

; ==================================================================================================
;  Misc functions
; ==================================================================================================
Func _TCHGetRandomString($nLength)
    Local $sString = ""
    For $i = 0 To $nLength - 1
        $sString &= Chr(Random(0x41, 0x5A))
    Next
    Return $sString
EndFunc

Func _TCHRestart($hWnd = 0, $bInvert = False)
    If $hWnd = 0 Then $hWnd = $TC_HWND
    If $hWnd = 0 Then
        MsgBox(48, "", _IntMsg("NoInstance"))
        Exit
    EndIf

    Local $bX64 = $TC_ARCH = 64 ; StringInStr(WinGetTitle($hWnd), "(x64)")
    If $bInvert Then $bX64 = Not $bX64

    WinClose($hWnd)
    WinWaitClose($hWnd)
    
    Run($TC_PATH & "\Totalcmd" & ($bX64 ? "64" : "") & ".exe")
    If @error Then MsgBox(48, "", _IntMsg("RunError"))
EndFunc

Func _TCHCheckCmdLine($nReq)
    If $CmdLine[0] < $nReq Then
        MsgBox(48, "", StringFormat(_IntMsg("NotEnoughParam"), $CmdLine[0], $nReq))
        Exit
    EndIf
EndFunc

; BETA, can be changed or removed
Func _TCHCheckCmdLineOpen($sDir = "", $sDefName = "", $sFilter = "All (*.*)", $hWnd = $TC_HWND)
    If $CmdLine[0] = 0 Then
        Local $sFileName = FileOpenDialog("Open file", $sDir, $sFilter, 1, $sDefName, $hWnd)
        If @error Then Exit
        Return $sFileName
    EndIf
EndFunc

; ==================================================================================================
; Keyboard management
; ==================================================================================================

Func _TCHGetKeyboardLayout($hWnd = _TCHGetScriptProcHandle())
    Local $ret = DllCall("user32.dll", "long", "GetWindowThreadProcessId", "hwnd", $hWnd, "ptr", 0)
          $ret = DllCall("user32.dll", "long", "GetKeyboardLayout", "long", $ret[0])
    Return "0000" & Hex($ret[0], 4)
EndFunc

Func _TCHSetKeyboardLayout($sLayoutID, $hWnd = _TCHGetScriptProcHandle())
    Local $WM_INPUTLANGCHANGEREQUEST = 0x50
    Local $ret = DllCall("user32.dll", "long", "LoadKeyboardLayout", "str", $sLayoutID, "int", 0)
    DllCall("user32.dll", "int", "SendMessage", "hwnd", $hWnd, _
                                                "int", $WM_INPUTLANGCHANGEREQUEST, _
                                                "int", 1, _
                                                "int", $ret[0])
EndFunc

Func _IsPressed($nHexKey, $vDLL = "user32.dll")
	Local $aReturn = DllCall($vDLL, "short", "GetAsyncKeyState", "int", $nHexKey)
	If @error Then Return SetError(@error, @extended, False)
	Return BitAND($aReturn[0], 0x8000) <> 0
EndFunc

; ==================================================================================================
; Clipboard management
; ==================================================================================================

; copy text with respect of current language layout
Func _TCHCopyClipboard($sText, $sLayoutID = "00000419")
    ; comment copied wrong instead
    _TCHSetKeyboardLayout($sLayoutID)
    ClipPut($sText)
EndFunc

; ==================================================================================================
; Commandline management
; ==================================================================================================

; Due the weird way that TC sends %N param on long files, handle commandline ourself
Func _TCHParseCmdLine($sCmdLine)
    Local $aParam[1] = [0]
    Local $qpos, $spos, $nDouble, $sCurParam
    While StringLen($sCmdLine) > 0
        $sCmdLine = StringStripWS($sCmdLine, 1)
        If StringLeft($sCmdLine, 1) = '"' Then
            $nDouble = 0
            if StringMid($sCmdLine, 2, 1) = '"' Then $nDouble = 1
            $qpos = StringInStr($sCmdLine, '"', 0, 2 + $nDouble)
            $sCurParam = StringMid($sCmdLine, 2 + $nDouble, $qpos - 2 - $nDouble)
            $aParam[0] += 1
            ReDim $aParam[$aParam[0]+1]
            $aParam[$aParam[0]] = $sCurParam
            $sCmdLine = StringTrimLeft($sCmdLine, $qpos + $nDouble)
            If StringLen($sCmdLine) = 1 Then
                ; rare condition that may occur when one redundant trailing qiote exists
                ExitLoop 
            Else
                ContinueLoop
            EndIf
        EndIf
        $spos = StringInStr($sCmdLine, ' ')
        If $spos > 0 Then
            $aParam[0] += 1
            ReDim $aParam[$aParam[0]+1]
            $aParam[$aParam[0]] = StringLeft($sCmdLine, $spos - 1)
            $sCmdLine = StringTrimLeft($sCmdLine, $spos)
        Else
            $aParam[0] += 1
            ReDim $aParam[$aParam[0]+1]
            $aParam[$aParam[0]] = $sCmdLine
            ExitLoop
        EndIf
        
    Wend
    Return $aParam
EndFunc

; ==================================================================================================
; Path/filename functions
; Those functions assume that strings that have "\" at the end - are dirs
; ==================================================================================================

; Return file extension, or empty string in case of no extension
Func _TCHExtractFileExt($sFile)
    ; regexp version is surprisingly twice (!) slower
    Local $nSPos = StringInStr($sFile, ".", 0, -1)
    If $nSPos > 0 Then
        If Not StringInStr($sFile, "\", 0, 1, $nSPos + 1) Then Return StringMid($sFile, $nSPos + 1)
    EndIf
    Return ""
EndFunc

; Return file name only, optionally with extension stripped
Func _TCHExtractFileName($sFile, $bStripExtension = False)
    ; regexp version is 1.5 to 2 times faster
    Local $sName = StringRegExpReplace($sFile, "(?:^.*\\|^)([^\\]*)$", "\1")
    If $bStripExtension Then
        ; one shot regexp for retrieve name without extension is slower then classic method
        Local $nSPos = StringInStr($sName, ".", 0, -1)
        If $nSPos > 0 Then Return StringLeft($sName, $nSPos - 1)
    EndIf 
    Return $sName
EndFunc

; Return file dir without trailing "\", empty string if no dir
Func _TCHExtractFileDir($sFile)
    Return StringLeft($sFile, StringInStr($sFile, "\", 0, -1) - 1)
EndFunc

; Return file dir with trailing "\", empty string if no dir
Func _TCHExtractFilePath($sFile)
    Return StringLeft($sFile, StringInStr($sFile, "\", 0, -1))
EndFunc

Func _TCHChangeFileExt($sFile, $sNewExt)
    Return StringRegExpReplace($sFile, "((?:^.+?\\|^)[^\\]+)\.[^\.\\]+$", "\1") & "." & $sNewExt
EndFunc

Func _TCHEnsureSlash($sFile, $bNeedSlash)
    Local $bHasSlash = StringRight($sFile, 1) = "\"
    If Not $bNeedSlash And $bHasSlash Then Return StringTrimRight($sFile, 1)
    If $bNeedSlash And Not $bHasSlash Then Return $sFile & "\"
    Return $sFile
EndFunc

; working dir only required when expanding relative paths
Func _TCHGetFullPath($sFileName, $sWorkDir = @ScriptDir)
    FileChangeDir($sWorkDir)
    Local $tBuffer = DllStructCreate("wchar sPath[32767]")
    Local $aRet = DllCall("kernel32.dll", "dword", "GetFullPathNameW", _
                                          "wstr", $sFileName, _
                                          "dword", DllStructGetSize($tBuffer), _
                                          "ptr", DllStructGetPtr($tBuffer), _
                                          "wstr*", "")
    Return $tBuffer.sPath
EndFunc

; ==================================================================================================
; Icon name management functions
; ==================================================================================================

; Return full path of current icon library
Func _TCHFindTCIconLibrary()
    Local Static $sIconLib = ""
    If $sIconLib <> "" Then Return $sIconLib
    Local $sIniLib = _TCHIniRead($TC_INI, "Buttonbar", "DefaultLib", "")
    If $sIniLib = "" Then
        $sIconLib = $TC_PATH & "\wciconex."
        If Not (FileExists($sIconLib & "dll") And FileExists($sIconLib & "inc")) Then
            $sIconLib = $TC_PATH & "\wcmicons."
        EndIf
        $sIconLib &= "dll"
    Else
        $sIconLib = $TC_PATH & "\" & $sIniLib & ".dll"
    EndIf
    Return $sIconLib
EndFunc

; split "Path_To_Icon,Index" path to array
; supported paths without index
; ensure that both array items are valid
; return array: Path, 0-Index, Negative 1-based index (for GUI functions)
Func _TCHSplitIconPath($sIconPath)
    Local $aIcon[3] = [$sIconPath, 0, -1]
    Local $aTempIcon = StringRegExp($sIconPath, "(.+?),(\d+)$", 1)
    If Not @error Then
        $aIcon[0] = $aTempIcon[0]
        $aIcon[1] = $aTempIcon[1]
        $aIcon[2] = -$aTempIcon[1]-1
    EndIf
    Return $aIcon
EndFunc

; Find icon in custom icons mappings
; return array: Path, 0-Index, Negative 1-based index (for GUI functions)
; or string as used in TC
; @extended: 2 - custom, 3 - wciconex redirected, 4 - internal command default mapping
Func _TCHFindCommandIcon($sCmdName, $bAsString = False)
    Local Static $sStdLib = _TCHFindTCIconLibrary()
    Local Static $sCustomLib = $OPTION("CustomIconLib")
;    Local Static $sCustomLibInc = _TCHChangeFileExt($sCustomLib, "inc")
    Local Static $sMapperConfig = $OPTION("IconMapperConfig")
    Local $nMapping = IniRead($sMapperConfig, "Custom", $sCmdName, $sIniError)
    If $nMapping = $sIniError Then 
        $nMapping = IniRead($sMapperConfig, "Wciconex", $sCmdName, $sIniError)
        If $nMapping <> $sIniError Then
            Local $nIconType = $IT_MAINLIB
        Else
            $nMapping = IniRead($sMapperConfig, "Default", $sCmdName, $sIniError)
            If $nMapping = $sIniError Then Return SetError(1, 0, "")
            Local $nIconType = $IT_MAINLIBDEF
        EndIf
        Local $aIcon[3] = [$sStdLib, $nMapping, -Number($nMapping)-1]
    Else
        Local $aIcon[3] = [$sCustomLib, $nMapping, -Number($nMapping)-1]
        Local $nIconType = $IT_CUSTOM
    EndIf
    If $bAsString Then Return SetError(0, $nIconType, $aIcon[0] & "," & $aIcon[1])
    Return SetError(0, $nIconType, $aIcon)
EndFunc

; BETA
; Finds script icon in order
; 1 - if compiled - first icon in exe
; 2 - icon file near script
; 3 - in meta-block
; 4 - find in custom icons mapping
; return array: Path, 0-Index, Negative 1-based index (for GUI functions)
; or string as used in TC
Func _TCHGetScriptIcon($sScriptFile = @ScriptFullPath, $bAsString = False)
    Local $aIcon[3] = [@AutoItExe, 0, -1]
    If @Compiled Then
        $aIcon[0] = @ScriptFullPath
        Return SetExtended(1, $bAsString ? $aIcon[0] & "," & $aIcon[1] : $aIcon)
    EndIf

    ; simple method - icon with the same name as script
    ; this method is for public source and for testing only
    Local $sIconName = _TCHChangeFileExt(@ScriptFullPath, "ico")
    If FileExists($sIconName) Then
        $aIcon[0] = $sIconName
        Return SetExtended(3, $bAsString ? $aIcon[0] & "," & $aIcon[1] : $aIcon)
    EndIf 

    Local $sIcon = IniRead($sScriptFile, "Info", "Icon", "")

    If $sIcon = "" Then
        ; Try to find custom icon
        Local $sCmdName = "em_" & _TCHExtractFileName($sScriptFile, True)
        Local $aCmdIcon = _TCHFindCommandIcon($sCmdName)
        If @error Then Return SetExtended(2, $aIcon)
        Switch @extended
            Case 2 ; get custom
                $aIcon[0] = $OPTION("CustomIconLib")
                $aIcon[1] = IniRead($OPTION("IconMapperConfig"), "Custom", $sCmdName, 0)
                $aIcon[2] = -Number($aIcon[1])-1
                Return $bAsString ? $aIcon[0] & "," & $aIcon[1] : $aIcon
            Case 3, 4 ; get mapped
                Return $bAsString ? $aCmdIcon[0] & "," & $aCmdIcon[1] : $aCmdIcon
        EndSwitch
    EndIf

    ; get explicit
    $aIcon = _TCHSplitIconPath($sIcon)
    $aIcon[0] = _TCHExpandEnv($aIcon[0])
    Return $aIcon
EndFunc

Func _TCHSetGUIDefaultIcon()
    Local Static $aIcon = _TCHGetScriptIcon()
    GUISetIcon($aIcon[0], $aIcon[2])
EndFunc

; ==================================================================================================
; Windows/Messages/GUI functions
; ==================================================================================================

; Standard versions of SendMessage/PostMessage
Func _TCHSendMessage($hWnd, $nMsg, $wParam = 0, $lParam = 0, $nResult = 0, $wParamType = "wparam", $lParamType = "lparam", $sReturnType = "lresult")
	Local $aResult = DllCall("user32.dll", "lresult", "SendMessageW", "hwnd", $hWnd, "uint", $nMsg, $wParamType, $wParam, $lParamType, $lParam)
	If @error Then Return SetError(@error, @extended, "")
	If $nResult >= 0 And $nResult <= 4 Then Return $aResult[$nResult]
	Return $aResult
EndFunc

Func _TCHPostMessage($hWnd, $nMsg, $wParam = 0, $lParam = 0, $nResult = 0, $wParamType = "wparam", $lParamType = "lparam", $sReturnType = "lresult")
	Local $aResult = DllCall("user32.dll", $sReturnType, "PostMessageW", "hwnd", $hWnd, "uint", $nMsg, $wParamType, $wParam, $lParamType, $lParam)
	If @error Then Return SetError(@error, @extended, "")
	If $nResult >= 0 And $nResult <= 4 Then Return $aResult[$nResult]
	Return $aResult
EndFunc

; Shortcut versions
Func _TCHSendMsg($vMsg, $hWnd = $TC_HWND)
    If IsString($vMsg) Then $vMsg = _IntCmd($vMsg)
    Return _TCHSendMessage($hWnd, $TC_SENDMESSAGE, $vMsg, 0)
EndFunc

Func _TCHPostMsg($vMsg, $hWnd = $TC_HWND)
    If IsString($vMsg) Then $vMsg = _IntCmd($vMsg)
    Return _TCHPostMessage($hWnd, $TC_SENDMESSAGE, $vMsg, 0)
EndFunc

Func _TCHRequestInfo($nMsg, $hWnd = $TC_HWND)
    Return _TCHSendMessage($hWnd, $TC_INFOREQUEST, $nMsg, 0)
EndFunc

; Executing user command
Func _TCHExecEMCommand($sCommand, $hTrgWnd = $TC_HWND)
    _TCHSendDataRequest("EM", $sCommand, $hTrgWnd)
EndFunc

Func _TCHSendDataRequest($sReqType, $sCommand, $hTrgWnd = $TC_HWND, $hSrcWnd = 0)
    Local Const $WM_COPYDATA = 0x004A
    Local $aChars = StringSplit($sReqType, "")
    If $aChars[0] <> 2 Then Return SetError(1)
    Local $pCommand = DllStructCreate("char[" & StringLen($sCommand) & "]") ; always ANSI
    DllStructSetData($pCommand, 1, $sCommand)
    Local $pCopyData = DllStructCreate("ulong_ptr;dword;ptr")
    DllStructSetData($pCopyData, 1, Asc($aChars[1]) + 256 * Asc($aChars[2]))
    DllStructSetData($pCopyData, 2, DllStructGetSize($pCommand))
    DllStructSetData($pCopyData, 3, DllStructGetPtr($pCommand))
    DllCall("user32.dll", "int", "SendMessage", "hwnd", $hTrgWnd, "int", $WM_COPYDATA, "wparam", $hSrcWnd, "lparam", DllStructGetPtr($pCopyData))
EndFunc

Func _TCHLockWindowUpdate($hWnd, $bLock)
    _TCHSendMessage($hWnd, 0x000B, $bLock ? 0 : 1) ; WM_SETREDRAW
EndFunc

; GUISetState works only with internal GUI
Func _TCHSetParentModal($hWnd, $bModal = True)
    WinSetState($hWnd, "", $bModal ? @SW_DISABLE : @SW_ENABLE)
EndFunc

Func _TCHTrayMessage($sMessage, $nDelay = 5000, $sTitle = _IntMsg("Message"))
    TrayTip($sTitle, $sMessage, 1, 1 + 16)
    Sleep($nDelay)
EndFunc

; Some functions require script window handle (Sender window for Send functons, context menu etc.)
Func _TCHGetScriptProcHandle()
    Local $sTitle = "ScriptWinRandomTitle" & _TCHGetRandomString(32)
    AutoItWinSetTitle($sTitle)
    Return WinGetHandle($sTitle)
EndFunc

; ==================================================================================================
; Environment variable functions
; ==================================================================================================

Func _TCHAddEnv($sText)
    ; Only first found will be replaced
    Local $aEnvList[] = ["COMMANDER_PATH","SYSTEMROOT","WINDIR","PROGRAMFILES","COMMONPROGRAMFILES", _
                           "USERPROFILE","ALLUSERSPROFILE","APPDATA"]
    For $i = 0 To UBound($aEnvList) - 1
        $sText = StringReplace($sText, EnvGet($aEnvList[$i]), "%" & $aEnvList[$i] & "%")
        If @extended > 0 Then Return $sText
    Next
    Return $sText
EndFunc

Func _TCHExpandEnv($sText)
    Local $sValue
    Local $aResult = StringRegExp($sText, "%(\w+)%", 3)
    If IsArray($aResult) Then
        For $i = 0 To UBound($aResult)-1
            $sValue = EnvGet($aResult[$i])
            If $sValue = "" Then ContinueLoop
            $sText = StringReplace($sText, "%" & $aResult[$i] & "%", $sValue)
        Next
    EndIf
    Return $sText
EndFunc

Func _TCHGetEditor($sExt)
    Local $sEditor = _TCHExpandEnv(_TCHIniRead($TC_INI, "Configuration", "Editor", "notepad.exe"))
    If StringRight($sEditor, 2) = "%1" Then $sEditor = StringStripWS(StringTrimRight($sEditor, 2), 3)
    ; this is NOT work directly for TC Edit Redirector due it require parent TC window
    If StringInStr($sEditor, "tcer.exe") Then 
        Local $sTcerIni = _TCHChangeFileExt($sEditor, "ini")
        Local $aEditors = IniReadSection($sTcerIni, "Extensions")
        Local $sDefEditor = IniRead($sTcerIni, "DefaultProgram", "FullPath", "")
        Local $aExts
        For $i = 1 To $aEditors[0][0]
            $aExts = StringSplit($aEditors[$i][0], ",")
            For $j = 1 To $aExts[0]
                If $aExts[$j] = $sExt Then Return _TCHExpandEnv(IniRead($sTcerIni, $aEditors[$i][1], "FullPath", $sDefEditor))
            Next
        Next
        Return _TCHExpandEnv($sDefEditor)
    EndIf 
    Return $sEditor
EndFunc

; ==================================================================================================
;  File functions
; ==================================================================================================
Func _TCHFileWriteReplace($sFileName, $sText, $nEncoding = 0, $bBackup = False)
    If $bBackup Then FileCopy($sFileName, _TCHExtractFilePath($sFileName) & _TCHExtractFileName($sFileName) & ".bak", 1)
    Local $hFile = FileOpen($sFileName, BitOR(2, $nEncoding))
    FileWrite($hFile, $sText)
    FileClose($hFile)
EndFunc

Func _TCHFileLoadList($sPath, $sFileMask = "*", $bFullPath = False, $bIncludeFolders = False)
    $sPath = _TCHEnsureSlash($sPath, True)
    Local $hSearch = FileFindFirstFile($sPath & $sFileMask)
    Local $sFile, $sOutput, $sOutPath = $bFullPath ? $sPath : ""
    While 1
        $sFile = FileFindNextFile($hSearch)
        If @error Then ExitLoop
        If @extended And Not $bIncludeFolders Then ContinueLoop
        $sOutput &= $sOutPath & $sFile & "|"
    WEnd
    FileClose($hSearch)
    Return StringSplit(StringTrimRight($sOutput, 1), "|", 2)
EndFunc

; ==================================================================================================
; Continious lists
; ==================================================================================================
; This function read "continious" lists of TC to 2D array
; sKeysFormat - string of key names, separated by "|". As counter substitute using @ char.
; $nStartCount - zero-based index of element from which list is read
; Example for reading WDX plugins list:
; @|@_detect will read keys: 1 and 1_detect, 2 and 2_detect etc.
; First item in list is mandatory, it's used for checking group presence
; ==================================================================================================
Func _TCHIniReadConList($sIniName, $sSection, $sKeysFormat, $nStartCount = 0)
    Local $nNewIdx, $sCurVal
    Local $aKeyNames = StringSplit($sKeysFormat, "|")
    Local $nKeys = $aKeyNames[0]
    Local $aResult[1][$nKeys] = [[0]]
    While 1
        $sCurVal = _TCHIniRead($sIniName, $sSection, StringReplace($aKeyNames[1], "@", $aResult[0][0]+$nStartCount), $sIniError)
        If $sCurVal <> $sIniError Then
            $aResult[0][0] += 1
            $nNewIdx = $aResult[0][0]
            ReDim $aResult[$nNewIdx+1][$nKeys]
            $aResult[$nNewIdx][0] = $sCurVal
            For $i = 2 To $aKeyNames[0]
                $sCurVal = _TCHIniRead($sIniName, $sSection, StringReplace($aKeyNames[$i], "@", $aResult[0][0]-1+$nStartCount), $sIniError)
                If $sCurVal <> $sIniError Then $aResult[$nNewIdx][$i-1] = $sCurVal
            Next
        Else
            Return $aResult
        EndIf
    Wend
EndFunc

; ==================================================================================================
; This function delete item from "continious" ini-lists
; sKeysFormat - string of key names, separated by "|". As counter substitute using @ char.
; $nIndex - zero-based index of element being deleted
; Example for reading WDX plugins list:
; @|@_detect will read keys: 1 and 1_detect, 2 and 2_detect etc.
; ==================================================================================================
Func _TCHIniRemoveConList($sIniName, $sSection, $sKeysFormat, $nIndex)
    Local $aList = _TCHIniReadConList($sIniName, $sSection, $sKeysFormat)
    $nIndex += 1 ; our lists are 1-based, legacy...
    If $nIndex > $aList[0][0] Then Return SetError(1, 0, -1)
    Local $aFields = StringSplit($sKeysFormat, "|", 2)
    Local $nCounter = -1
    For $i = 1 To $aList[0][0]
        If $i = $nIndex Then ContinueLoop ; skip item with index being deleting
        $nCounter += 1
        For $j = 0 To UBound($aFields)-1
            _TCHIniWrite($sIniName, $sSection, StringReplace($aFields[$j], "@", $nCounter), $aList[$i][$j])
        Next
    Next
    ; remove last group
    $nCounter += 1
    For $j = 0 To UBound($aFields)-1
        _TCHIniDelete($sIniName, $sSection, StringReplace($aFields[$j], "@", $nCounter))
    Next
EndFunc

; ==================================================================================================
; Redirection-aware ini functions
; Has the same interface as internal Autoit functions
; ==================================================================================================
Func _TCHIniGetRedirectName($sIniName, $sSection)
    Local $sBasePath = ""
    Local $sRedirect = IniRead($sIniName, $sSection, "RedirectSection", $sIniError)
    If $sRedirect <> $sIniError Then
        If StringInStr($sRedirect, "%") Then $sRedirect = _TCHExpandEnv($sRedirect)
        ; if path is local or UNC then return redirected path itself
        If StringInStr($sRedirect, ":") Or (StringLeft($sRedirect, 2) = "\\")  Then Return $sRedirect
        ; else assume redirect path is relative and extract source ini base path
        $sBasePath = StringRegExpReplace($sRedirect, "(.+\\)([^\\]+)", "\1")
        $sIniName = $sBasePath & $sRedirect
    EndIf
    Return $sIniName
EndFunc

Func _TCHIniDelete($sIniName, $sSection, $sKey = "")
    $sIniName = _TCHIniGetRedirectName($sIniName, $sSection)
    If $sKey = "" Then
        Return IniDelete($sIniName, $sSection)
    Else
        Return IniDelete($sIniName, $sSection, $sKey)
    EndIf
EndFunc

; clear section without deleting it
Func _TCHIniClearSection($sIniName, $sSection)
    $sIniName = _TCHIniGetRedirectName($sIniName, $sSection)
    Local $aKeys = IniReadSection($sIniName, $sSection)
    If UBound($aKeys) = 0 Then Return
    For $i = 1 To $aKeys[0][0]
        IniDelete($sIniName, $sSection, $aKeys[$i][0])
    Next
EndFunc

Func _TCHIniRead($sIniName, $sSection, $sKey, $sDefault = '')
    $sIniName = _TCHIniGetRedirectName($sIniName, $sSection)
    Return IniRead($sIniName, $sSection, $sKey, $sDefault)
EndFunc

Func _TCHIniReadSection($sIniName, $sSection)
    $sIniName = _TCHIniGetRedirectName($sIniName, $sSection)
    Return IniReadSection($sIniName, $sSection)
EndFunc

Func _TCHIniWrite($sIniName, $sSection, $sKey, $sValue)
    $sIniName = _TCHIniGetRedirectName($sIniName, $sSection)
    Return IniWrite($sIniName, $sSection, $sKey, $sValue)
EndFunc

; ==================================================================================================
; Button and bar functions
; ==================================================================================================
Func _TCHCreateButton($sCmd, $sParam = "", $sIcon = "", $sTitle = "", $sPath = "", $nShow = 0, $nCMCmd = -1)
    Local $sFormat = $OPTION("BarDataID") & "\r\n%s\r\n%s\r\n%s\r\n%s\r\n%s\r\n%d\r\n%d\r\n"
    Return StringFormat($sFormat, $sCmd, $sParam, $sIcon, $sTitle, $sPath, $nShow, $nCMCmd)
EndFunc

; ==================================================================================================
; Internal helper functions
; ==================================================================================================

; Get instance of Total Commander from which script ran (in order multiple copies runned)
Func __TCHGetActiveInstance()
    Local $sInst = EnvGet("COMMANDER_HWND") ; if set by Autorun, preferable
    If ($sInst <> "") And StringIsDigit($sInst) Then Return Hwnd($sInst)
    ; Fallback to "active" window method
    Local $aList = WinList("[CLASS:TTOTAL_CMD]")
    For $i = 1 To $aList[0][0]
        If WinActive($aList[$i][1]) Then Return $aList[$i][1]
    Next
    Return 0
EndFunc

; Try to get install path of TC (in priority: EnvVar, HKLM key, HKCU key)
Func __TCHGetPath()
    Local $sTCPath = EnvGet("COMMANDER_PATH")
    If $sTCPath = "" Then ; unlikely, legacy code below
        $sTCPath = RegRead($TC_REGKEY_SYSTEM, "InstallDir")
        If $sTCPath = "" Then
            $sTCPath = RegRead($TC_REGKEY_USER, "InstallDir")
            If $sTCPath = "" Then
                MsgBox(48, "", _IntMsg("NoGetPath"))
                Exit
            EndIf
        EndIf
    EndIf
    Return $sTCPath
EndFunc

Func __TCHGetArch()
    Local $nArch = EnvGet("COMMANDER_ARCH") ; if set by Autorun
    If $nArch = "" Then
        ; much less reliable, is there any other way to get?
        If StringInStr(WinGetTitle($TC_HWND), "(x64)") Then
            Return 64
        Else
            Return 32
        EndIf
    EndIf
    Return $nArch
EndFunc

Func __TCHGetLangID()
    Local $sLng = EnvGet("COMMANDER_LANG") ; if set by Autorun
    If $sLng = "" Then
        $sLng = _TCHIniRead($TC_INI, "Configuration", "LanguageIni", "WCMD_ENG.LNG")
        $sLng = StringRegExpReplace($sLng, "(?i).+[_-]([a-zA-Z]+)\.lng", "\1")
    EndIf
    Return $sLng
EndFunc

Func __TCHGetTypesLib()
    Local Static $sTypesLibrary = $TC_ICONS & "\" & IniRead(_Option("IconMapperConfig"), "Settings", "TypesLib", "Sources\Types.icl")
    Return $sTypesLibrary
EndFunc